"use strict";
/*
 * Deepkit Framework
 * Copyright (C) 2020 Deepkit UG
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.handleActiveSubject = exports.MessageSubject = void 0;
const rxjs_1 = require("rxjs");
const operators_1 = require("rxjs/operators");
const core_1 = require("./core");
const collection_1 = require("./collection");
const type_1 = require("@deepkit/type");
const core_2 = require("@deepkit/core");
class MessageSubject extends rxjs_1.Subject {
    constructor(connectionId, onReply) {
        super();
        this.connectionId = connectionId;
        this.onReply = onReply;
        this.disconnected = new rxjs_1.Subject();
    }
    setSendMessageModifier(modifier) {
        this.sendModifier = modifier;
    }
    /**
     * Sends a message to the server and returns a new MessageSubject.
     * If the connection meanwhile has been reconnected, and completed MessageSubject.
     */
    sendMessage(messageWithoutId) {
        if (!this.onReply) {
            throw new Error('No replier set');
        }
        if (this.sendModifier) {
            messageWithoutId = this.sendModifier(messageWithoutId);
        }
        return this.onReply(messageWithoutId);
    }
    error(err) {
        this.disconnected.complete();
        super.error(err);
    }
    complete() {
        this.disconnected.complete();
        super.complete();
    }
    async firstOrUndefinedThenClose() {
        if (this.closed) {
            return undefined;
        }
        return new Promise((resolve) => {
            this.pipe(operators_1.first()).subscribe((next) => {
                resolve(next);
            }, (error) => {
                resolve(undefined);
            }, () => {
                //complete
            }).add(() => {
                this.complete();
            });
        });
    }
    async firstThenClose() {
        return new Promise((resolve, reject) => {
            this.pipe(operators_1.first()).subscribe((next) => {
                resolve(next);
            }, (error) => {
                reject(error);
            }, () => {
                //complete
            }).add(() => {
                this.complete();
            });
        });
    }
}
exports.MessageSubject = MessageSubject;
function handleActiveSubject(activeSubject, resolve, reject, controller, name, entityState, options) {
    let returnValue;
    const subscribers = {};
    let subscriberIdCounter = 0;
    let streamBehaviorSubject;
    function deserializeResult(encoding, next) {
        return type_1.jsonSerializer.deserializeProperty(type_1.PropertySchema.fromJSON(encoding), next);
    }
    const sub = activeSubject.subscribe((reply) => {
        if (reply.type === 'type') {
            if (reply.returnType === 'subject') {
                if (options && options.useThisStreamBehaviorSubject) {
                    streamBehaviorSubject = options.useThisStreamBehaviorSubject;
                    streamBehaviorSubject.next(deserializeResult(reply.encoding, reply.data));
                }
                else {
                    streamBehaviorSubject = new core_1.StreamBehaviorSubject(deserializeResult(reply.encoding, reply.data));
                }
                //this behavior is not consistent, so we commented it out.
                // const reconnectionSub = activeSubject.reconnected.subscribe(() => {
                //     reconnectionSub.unsubscribe();
                //     this.stream(controller, name, args, {useThisStreamBehaviorSubject: streamBehaviorSubject});
                // });
                streamBehaviorSubject.addTearDown(async () => {
                    // reconnectionSub.unsubscribe();
                    //user unsubscribed the entity subject, so we stop syncing changes
                    await activeSubject.sendMessage({
                        name: 'subject/unsubscribe',
                        forId: reply.id,
                    }).firstOrUndefinedThenClose();
                    streamBehaviorSubject = undefined;
                    activeSubject.complete();
                    sub.unsubscribe();
                });
                resolve(streamBehaviorSubject);
            }
            if (reply.returnType === 'entity') {
                if (reply.item) {
                    if (!type_1.hasClassSchemaByName(reply.entityName)) {
                        throw new Error(`Entity ${reply.entityName} not known. (known: ${type_1.getKnownClassSchemasNames().join(',')})`);
                    }
                    const classType = type_1.getClassSchemaByName(reply.entityName).classType;
                    const subject = entityState.handleEntity(classType, reply.item);
                    subject.addTearDown(async () => {
                        //user unsubscribed the entity subject, so we stop syncing changes
                        await activeSubject.sendMessage({
                            name: 'entity/unsubscribe',
                            forId: reply.id,
                        }).firstOrUndefinedThenClose();
                        activeSubject.complete();
                        sub.unsubscribe();
                    });
                    resolve(subject);
                }
                else {
                    reject(new Error('Item not found'));
                }
            }
            if (reply.returnType === 'observable') {
                returnValue = new rxjs_1.Observable((observer) => {
                    const subscriberId = ++subscriberIdCounter;
                    subscribers[subscriberId] = observer;
                    activeSubject.sendMessage({
                        forId: reply.id,
                        name: 'observable/subscribe',
                        subscribeId: subscriberId
                    }).firstOrUndefinedThenClose();
                    return {
                        unsubscribe() {
                            activeSubject.sendMessage({
                                forId: reply.id,
                                name: 'observable/unsubscribe',
                                subscribeId: subscriberId
                            }).firstOrUndefinedThenClose();
                            //we need to keep the activeSubject alive,
                            //since the user could create new subscriptions.
                        }
                    };
                });
                resolve(returnValue);
            }
            if (reply.returnType === 'collection') {
                if (!type_1.hasClassSchemaByName(reply.entityName)) {
                    throw new Error(`Entity ${reply.entityName} not known. (known: ${type_1.getKnownClassSchemasNames().join(',')})`);
                }
                const classType = type_1.getClassSchemaByName(reply.entityName).classType;
                const collection = new collection_1.Collection(classType);
                if (reply.pagination.active) {
                    collection.pagination._activate();
                    collection.pagination.setItemsPerPage(reply.pagination.itemsPerPage);
                    collection.pagination.setTotal(reply.pagination.total);
                    collection.pagination.setPage(reply.pagination.page);
                    collection.pagination.setSort(reply.pagination.sort);
                }
                collection.pagination.setParameters(reply.pagination.parameters);
                collection.pagination.event.subscribe((event) => {
                    if (event.type === 'apply') {
                        activeSubject.sendMessage({
                            forId: reply.id,
                            name: 'collection/pagination',
                            sort: collection.pagination.getSort(),
                            parameters: collection.pagination.getParameters(),
                            page: collection.pagination.getPage(),
                            itemsPerPage: collection.pagination.getItemsPerPage(),
                        }).firstOrUndefinedThenClose();
                    }
                });
                returnValue = collection;
                collection.addTeardown(async () => {
                    for (const entitySubject of core_2.each(collection.entitySubjects)) {
                        entitySubject.unsubscribe();
                    }
                    //collection unsubscribed, so we stop syncing changes
                    await activeSubject.sendMessage({
                        forId: reply.id,
                        name: 'collection/unsubscribe'
                    }).firstOrUndefinedThenClose();
                    returnValue = undefined;
                    activeSubject.complete();
                    sub.unsubscribe();
                });
                //do not resolve yet, since we want to wait until the collection has bee populated.
            }
        }
        if (reply.type === 'next/json') {
            resolve(deserializeResult(reply.encoding, reply.next));
        }
        if (reply.type === 'next/observable') {
            if (subscribers[reply.subscribeId]) {
                subscribers[reply.subscribeId].next(deserializeResult(reply.encoding, reply.next));
            }
        }
        if (reply.type === 'next/subject') {
            if (streamBehaviorSubject) {
                if (!streamBehaviorSubject.isUnsubscribed()) {
                    streamBehaviorSubject.next(deserializeResult(reply.encoding, reply.next));
                }
            }
        }
        if (reply.type === 'append/subject') {
            if (streamBehaviorSubject) {
                if (!streamBehaviorSubject.isUnsubscribed()) {
                    const append = deserializeResult(reply.encoding, reply.append);
                    streamBehaviorSubject.append(append);
                }
            }
        }
        if (reply.type === 'next/collection') {
            entityState.handleCollectionNext(returnValue, reply.next);
            resolve(returnValue);
        }
        if (reply.type === 'complete') {
            if (returnValue instanceof collection_1.Collection) {
                returnValue.complete();
                returnValue = undefined;
                activeSubject.complete();
                sub.unsubscribe();
            }
            if (streamBehaviorSubject) {
                streamBehaviorSubject.complete();
                streamBehaviorSubject = undefined;
                activeSubject.complete();
                sub.unsubscribe();
            }
            //activeSubject.complete();
        }
        if (reply.type === 'error') {
            const error = core_1.getUnserializedError(reply.entityName, reply.error, reply.stack, `action ${controller}.${name}`);
            activeSubject.complete();
            sub.unsubscribe();
            if (returnValue instanceof collection_1.Collection) {
                returnValue.error(error);
                returnValue = undefined;
            }
            else if (streamBehaviorSubject) {
                streamBehaviorSubject.error(error);
                streamBehaviorSubject = undefined;
            }
            else {
                reject(error);
            }
        }
        if (reply.type === 'error/observable') {
            const error = core_1.getUnserializedError(reply.entityName, reply.error, reply.stack, `action ${controller}.${name}`);
            if (subscribers[reply.subscribeId]) {
                subscribers[reply.subscribeId].error(error);
            }
            delete subscribers[reply.subscribeId];
            //we dont end the subject since from the observable we could create new subscriptions
            //activeSubject.complete();
        }
        if (reply.type === 'complete/observable') {
            if (subscribers[reply.subscribeId]) {
                subscribers[reply.subscribeId].complete();
            }
            delete subscribers[reply.subscribeId];
            //we dont end the subject since from the observable we could create new subscriptions
            // activeSubject.complete();
        }
    }, (error) => {
        reject(error);
    }, () => {
    });
}
exports.handleActiveSubject = handleActiveSubject;
//# sourceMappingURL=client.js.map